function [energies,A] = dmrg_run(C,D,PHYS,nsweeps);
% 
% run DMRG optimization. Parameters:
%   
%   C: Hamiltonian MPO tensors
%      C{k}(l,r,down,up) with left, right, down up index
%
%   D: list of bond dimensions
%   PHYS: list of physical dimenions
%   nsweeps: number of sweeps
%
% return values: 
%   energies: energies in each step (e.g. for convergence)
%   A: optimal MPS tensors

energies = [];    % used to store energies in each step (for plotting)
N=length(C);      % number of sites

% we want to keep any bond dimension <= than the physical times other bond
% dimension of the adjacent tensors: choosing larger numbers does not
% increase the class of states we can describe, but would make the SVD for
% the canonical form more cumbersome. This happens now:
D{1} = min([D{1} PHYS{1}]);
D{N-1} = min([D{N-1} PHYS{N}]);
found = true;
while (found)
    found = false;
    for s=1:(N-2)
	if D{s}>D{s+1}*PHYS{s+1} found=true; end;
	if D{s+1}>D{s}*PHYS{s+1} found=true; end;
	D{s} = min([D{s} D{s+1}*PHYS{s+1}]);
	D{s+1} = min([D{s+1} D{s}*PHYS{s+1}]);
    end;
end;




A = init_rand_MPS(N,D,PHYS); % initialize MPS with random values. 
        % structure: A{k}(l,p,r) with left, physical, right.
        % N: length of chain
        % D{k}: bond dimensions (right of of k)
	% PHYS: list of physical dimensions

L{1}=1;   % L{k}(ket,mpo,bra): left contraction of Ham. <k
	  % trivial contraction of MPO left of 1
R{N}=1;   % R{k}(ket,mpo,bra): right contraction Ham. >k
	  % trivial contraction of MPO right of N

% ensure correct gauge condition (k0=1) and pre-compute R's.
for s=N:-1:2
    [A{s-1}, A{s}] = move_wsite_left(A{s-1},A{s}); % gauge condition
    R{s-1} = apply_MPO_Top_right(R{s},A{s},C{s});  % R
end;
A{1} = A{1}/norm(reshape(A{1},[prod(size(A{1})) 1]));


% perform nsweep sweeps
for sweep=1:nsweeps
    
    % part 1: right sweep
    tic;
    fprintf('Right sweep #%03i ... ',sweep);
    for s=2:N % move from s-1 -> s
	% update MPS tensor and L
	[A{s-1}, A{s}] = move_wsite_right(A{s-1},A{s});  % update gauge
	L{s} = apply_MPO_Top_left(L{s-1},A{s-1},C{s-1}); % update L{s}
	% optimize energy 
	[A{s},energy] = optimize_tensor(C{s},L{s},R{s});
	% store energy value	
	energies = [energies energy];
    end;

    fprintf('Left sweep #%03i ... ',sweep);
    % part 2: left sweep
    for s=(N-1):-1:1 % move from s+1 -> s
	% update MPS tensor and L
	[A{s},A{s+1}] = move_wsite_left(A{s},A{s+1});     % update gauge
	R{s} = apply_MPO_Top_right(R{s+1},A{s+1},C{s+1}); % update R{s}
	% optimize energy	
	[A{s},energy] = optimize_tensor(C{s},L{s},R{s});
	% store energy value	
	energies = [energies energy];
    end;
    fprintf('Done. Energy: %14.10f. Time: %0.1f secs\n',energy(end),toc);

end; % end of sweeps

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [BL,BR] = move_wsite_left(AL,AR);
% move working site to the left: make BR an isometry 
% and absorb the remaining part in BL

[U,D,V]=svd(reshape(AR,[size(AR,1) size(AR,2)*size(AR,3)]),'econ');
% this returns the "minimal dimension" SVD with U*D*V' = AR
if min(diag(D))<1e-13
    % safety check: Small singular values are potentially problematic.
    % NOTE: This could be fixed by "dynamically" cutting the bond
    % dimension if those become too small.
    warning('Close to singular conditioning number.');
end;
BR = reshape(V',[size(AR,1) size(AR,2) size(AR,3)]); % V' becomes the new BR
BL = ncon({AL,U*D},{[-1 -2 1],[1 -3]});   % absorb the remaining part in BL

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [BL,BR] = move_wsite_right(AL,AR);
% move working site to the right: make BL and isometry
% see comments in function above

[U,D,V]=svd(reshape(AL,[size(AL,1)*size(AL,2) size(AL,3)]),'econ');
if min(diag(D))<1e-13
    warning('Close to singular conditioning number.');
end;
BL = reshape(U,[size(AL,1) size(AL,2) size(AL,3)]);
BR = ncon({D*V',AR},{[-1 1],[1 -2 -3]});

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Lnew = apply_MPO_Top_left(Lold,A,C);
% update the left contraction of MPS + MPO in between
Lnew = ncon({Lold,A,C,conj(A)},{[1 3 5],[1 2 -1],[3 -2 2 4],[5 4 -3]});

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Rnew = apply_MPO_Top_right(Rold,A,C);
% update the right contraction of MPS + MPO in between
Rnew = ncon({Rold,A,C,conj(A)},{[1 3 5],[-1 2 1],[-2 3 2 4],[-3 4 5]});

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [A,energy]=optimize_tensor(C,L,R);

% find the optimal tensor (minimal energy) for MPO C and boundaries L and R
% A: optimal tensor. energy: optimal energy

Asize = [size(L,1) size(C,3) size(R,1)];  % size of tensor A

if false
    % variant 1: build the matrix acting on A and fully diagonalize it
    % should scale like D^6
    M = reshape(ncon({L,C,R},{[-4 1 -1],[1 2 -5 -2],[-6 2 -3]}),[prod(Asize) prod(Asize)]);
    [U,D]=eig(M);		    % diagonalize the matrix
    [energy,pos]=min(real(diag(D)));  % find smallest eigenvalue
    A = reshape(U(:,pos)/norm(U(:,pos)),Asize);  
        % shape into a tensor
    %
end;

if true
    % variant 2: use a Krylov method
    opts.isreal=false;
    [U,D]=eigs(...
	    @(x)(reshape(...
		ncon({L,C,R,reshape(x,Asize)},{[1 3 -1],[3 5 2 -2],[4 5 -3],[1 2 4]}),...
		[prod(Asize) 1])),...
	    prod(Asize),1,'sr',opts);
    energy=D;
    A = reshape(U/norm(U),Asize);
end; 
	    



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A=init_rand_MPS(N,D,PHYS);
% produce a random initial MPS
        % structure: A{k}(l,p,r) with left, physical, right.
        % N: length of chain
        % D{k}: bond dimensions (right of of k)
	% PHYS: list of physical dimensions
% first site is differerent
A{1}=randn(1,PHYS{1},D{1})+i*randn(1,PHYS{1},D{1}); 
% all the middle sites
for s=2:(N-1)
    A{s}=randn(D{s-1},PHYS{s},D{s})+i*randn(D{s-1},PHYS{s},D{s}); 
end;
% last site is different 
A{N}=randn(D{N-1},PHYS{N},1)+i*randn(D{N-1},PHYS{N},1); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

